# -*- coding: utf-8 -*-
from __future__ import division, print_function
import numpy as np
from pambox.speech.sepsm import Sepsm
from collections import namedtuple
from six.moves import zip


CENTER_F = (63, 80, 100, 125, 160, 200, 250, 315, 400, 500, 630, 800, 1000,
            1250, 1600, 2000, 2500, 3150, 4000, 5000, 6300, 8000)
HT_DIFFUSE = (37.5, 31.5, 26.5, 22.1, 17.9, 14.4, 11.4, 8.4, 5.8, 3.8, 2.1,
              1.0, 0.8, 1.9, 0.5, -1.5, -3.1, -4.0, -3.8, -1.8, 2.5, 6.8)
MODF = (1., 2., 4., 8., 16., 32., 64.)
FS = 22050.0


class BinauralSepsm(Sepsm):
    """Binaural implementation of the sEPSM model. """

    def __init__(self, fs=22050, cf=CENTER_F, modf=MODF, downsamp_factor=10,
                 noise_floor=0.01, snr_env_limit=0.001):
        """@todo: to be defined1. """
        Sepsm.__init__(self, fs, cf, modf, downsamp_factor, noise_floor,
                       snr_env_limit)

    def predict(self, clean, mixture, noise):
        """Predict intelligibility.

        :clean: @todo
        :mixture: @todo
        :noise: @todo
        :returns: @todo

        """
        binaural_res = [Sepsm.predict(self, c, m, n)
                        for c, m, n in zip(clean, mixture, noise)]
        Ears = namedtuple('Ears', ['left', 'right'])
        ears_res = Ears(*binaural_res)

        # Better ear (BE)
        be_matrix = self._better_ear(binaural_res)
        be_mask = self._get_bands_above_threshold(binaural_res)
        be_snr_env = Sepsm._optimal_combination(self, be_matrix, be_mask)

        res = namedtuple('Results', ['be_snr_env', 'ears', 'be_matrix',
                                     'be_above_thres'])
        res.be_snr_env = be_snr_env
        res.be_matrix = be_matrix
        res.be_above_thres = be_mask
        res.ears = ears_res
        return res  # Results for each ear's sEPSM model.

    def _better_ear(self, res):
        """Return the better-ear SNRenv for bands above threshold only.

        :param res: Results from Sepsm.predict()
        :type res: namedtuple
        :returns: List of 3 ndarray, for clean speech, mixture, and noise which
            contains the better ear SNRenv.

        """
        be_snr_env = np.zeros_like(res[0].snr_env_matrix)
        for ii in range(2):
            idx = res[ii].bands_above_thres_idx
            be_snr_env[idx] = \
                np.maximum(be_snr_env[idx], res[ii].snr_env_matrix[idx])
        return be_snr_env

    def _get_bands_above_threshold(self, res):
        """Return a mask for bands above threshold in at least one ear.

        :param res: Results return by the Sepsm.predict() function.
        :type res: structured array.
        :returns: ndarray of booleans -- The mask, it has the same dimensions
            as res.snr_env_matrix.

        """
        at = {}
        for ii in range(2):
            at[ii] = np.zeros_like(res[ii].snr_env_matrix)
            at[ii][res[ii].bands_above_thres_idx] = 1
        all_above_thres = np.logical_and(at[0], at[1])
        return all_above_thres
